"""
======
消息
======
实现了常用的基本消息 ``Message`` 和消息字段 ``MessageSegment`` 模型供适配器使用。
适配器开发者可以根据需要实现此模块中消息类的子类或定义与此不同的消息类型，但建议若可行的话应尽量使用此模块中消息类的子类。
"""
import json
import dataclasses
from copy import copy, deepcopy
from dataclasses import dataclass, field
from typing import Any, Dict, Mapping, Union, TypeVar, Iterable

T_Message = TypeVar('T_Message', bound='Message')
T_MessageSegment = TypeVar('T_MessageSegment', bound='MessageSegment')


class Message(list):
    """
    消息。
    本类是 ``list`` 的子类，同时重写了 ``__init__()`` 方法，可以直接处理 str, Mapping, Iterable[Mapping], T_MessageSegment, T_Message。
    其中 str 的支持需要适配器开发者重写 ``_construct()`` 方法实现，若开发者实现了 MessageSegment 的子类也需要重写 ``_construct()`` 方法。
    本类重写了 ``+`` 和 ``+=`` 运算符，可以直接和 Message, MessageSegment 等类型的对象执行取和运算。
    """

    def __init__(self,
                 message: Union[str, None, Mapping, Iterable[Mapping], T_MessageSegment, T_Message, Any] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        if message is None:
            return
        elif isinstance(message, Message):
            self.extend(message)
        elif isinstance(message, MessageSegment):
            self.append(message)
        else:
            self.extend(self._construct(message))

    @classmethod
    def __get_validators__(cls):
        """pydantic自定义校验器"""
        yield cls._validate

    @classmethod
    def _validate(cls, value):
        return cls(value)

    @staticmethod
    def _construct(msg: Union[str, Mapping, Iterable[Mapping], Any]) -> Iterable[T_MessageSegment]:
        """
        用于将 str, Mapping, Iterable[Mapping] 等类型转换为 MessageSegment。
        用于pydantic数据解析和方便用户使用，子类应重写此方法以支持 str 及支持新的消息字段。

        :param msg: 要解析为 MessageSegment 的数据。
        :return: MessageSegment 生成器。
        :rtype: Iterable[T_MessageSegment]
        """
        if isinstance(msg, Mapping):
            yield MessageSegment(msg['type'], msg.get('data') or {})
        elif isinstance(msg, Iterable) and not isinstance(msg, str):
            for seg in msg:
                yield MessageSegment(seg['type'], seg.get('data') or {})
        elif isinstance(msg, str):
            raise NotImplementedError
        return

    def __repr__(self) -> str:
        return 'Message:[{}]'.format(','.join(map(lambda x: repr(x), self)))

    def __str__(self) -> str:
        return ''.join(map(lambda x: str(x), self))

    def __contains__(self, item) -> bool:
        if isinstance(item, str):
            return item in str(self)
        return super().__contains__(item)

    def __add__(self: T_Message, other: Union[T_Message, T_MessageSegment, str]) -> T_Message:
        return self.__class__(self).__iadd__(other)

    def __radd__(self: T_Message, other: Union[T_Message, T_MessageSegment, str]) -> T_Message:
        return self.__class__(other).__iadd__(self)

    def __iadd__(self: T_Message, other: Union[T_Message, T_MessageSegment, str]) -> T_Message:
        if isinstance(other, type(self)):
            self.extend(other)
        if isinstance(other, MessageSegment):
            self.append(other)
        elif isinstance(other, str):
            self.extend(self._construct(other))
        else:
            raise TypeError(f"unsupported operand type(s) for +: '{type(self)!r}' and '{type(other)!r}'")
        return self

    def copy(self) -> T_Message:
        """
        返回自身的浅复制。

        :return: 自身的浅复制。
        :rtype: T_Message
        """
        return self.__class__(self)

    def deepcopy(self) -> T_Message:
        """
        返回自身的深复制。

        :return: 自身的深复制。
        :rtype: T_Message
        """
        return deepcopy(self)

    def startswith(self, prefix: str, start=None, end=None) -> bool:
        """
        实现类似字符串的 ``startswith()`` 方法。

        :param prefix: 前缀。
        :param start: 开始检查位置。
        :param end: 停止检查位置。
        :return: 检查结果。
        :rtype: bool
        """
        return str(self).startswith(prefix, start, end)

    def endswith(self, suffix: str, start=None, end=None) -> bool:
        """
        实现类似字符串的 ``endswith()`` 方法。

        :param suffix: 后缀。
        :param start: 开始检查位置。
        :param end: 停止检查位置。
        :return: 检查结果。
        :rtype: bool
        """
        return str(self).endswith(suffix, start, end)

    def replace(self,
                old: Union[str, T_MessageSegment],
                new: Union[str, T_MessageSegment, None],
                count: int = -1) -> T_Message:
        """
        实现类似字符串的 ``replace()`` 方法。
        当 ``old`` 为 str 类型时，``new`` 也必须是 str ，本方法将仅对 ``type`` 为 ``text`` 的消息字段进行处理。
        当 ``old`` 为 MessageSegment 类型时，``new`` 可以是 MessageSegment 或 None，本方法将对所有消息字段进行处理，并替换符合条件的消息字段。None 表示删除匹配到的消息字段。

        :param old: 被匹配的字符串或消息字段。
        :param new: 被替换为的字符串或消息字段。
        :param count: 替换的次数。
        :return:
        """
        temp_msg = self.deepcopy()
        if not (type(old) == type(new) == str) and \
                not (type(old) == MessageSegment and (type(new) == MessageSegment or new is None)):
            raise ValueError()
        if type(old) == str:
            for index, item in enumerate(temp_msg):
                if count == 0:
                    break
                if item.type == 'text' and old in item.data['text']:
                    if count == -1:
                        temp_msg[index].data['text'] = item.data['text'].replace(old, new)
                    else:
                        temp = item.data['text'].count(old)
                        temp_msg[index].data['text'] = item.data['text'].replace(old, new, min(temp, count))
                        if count <= temp:
                            count = 0
                        else:
                            count -= temp
        else:
            if new is None:
                temp_msg = Message(filter(lambda x: x != old, temp_msg))
            for index, item in enumerate(temp_msg):
                if count == 0:
                    break
                if item == old:
                    temp_msg[index] = new
        return temp_msg


@dataclass
class MessageSegment(Mapping):
    """
    消息字段。
    本类实现了所有映射类型的方法，这些方法全部是对 ``data`` 属性的操作。
    本类重写了 ``+`` 和 ``+=`` 运算符，可以直接和 Message, MessageSegment 等类型的对象执行取和运算，返回 Message 对象。
    """
    type: str
    """
    消息字段类型。
    """
    data: Dict[str, Any] = field(default_factory=lambda: {})
    """
    消息字段内容。
    """

    def __str__(self) -> str:
        return str(self.data)

    def __repr__(self) -> str:
        return f'MessageSegment<{self.type}>:{str(self)}'

    def __getitem__(self, key) -> Any:
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self):
        yield from self.data.__iter__()

    def __contains__(self, key) -> bool:
        return key in self.data

    def __eq__(self, other: T_MessageSegment) -> bool:
        return self.type == other.type and self.data == other.data

    def __ne__(self, other: T_MessageSegment) -> bool:
        return not self.__eq__(other)

    def __add__(self, other):
        return Message(self) + other

    def __radd__(self, other):
        return Message(other) + self

    def get(self, key: str, default=None):
        return getattr(self.data, key, default)

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    def copy(self) -> T_MessageSegment:
        """
        返回自身的浅复制。

        :return: 自身的浅复制。
        :rtype: T_MessageSegment
        """
        return copy(self)

    def deepcopy(self) -> T_MessageSegment:
        """
        返回自身的深复制。

        :return: 自身的深复制。
        :rtype: T_MessageSegment
        """
        return deepcopy(self)


class DataclassEncoder(json.JSONEncoder):
    """
    用于解析 MessageSegment 的 JSONEncoder 类。
    """

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)



"""
============
协议适配器
============
所有协议适配器都必须继承自 ``AbstractAdapter`` 基类。
"""
import time
import asyncio
from abc import ABC, abstractmethod
from functools import wraps, partial
from typing import Awaitable, List, Callable, TypeVar, Union, Optional, TYPE_CHECKING

from alicebot.log import logger

if TYPE_CHECKING:
    from alicebot import Bot
    from alicebot.event import T_Event

T_Adapter = TypeVar('T_Adapter', bound='AbstractAdapter')


class AbstractAdapter(ABC):
    """
    协议适配器基类。
    """
    name: str
    """
    适配器的名称。
    """
    bot: 'Bot'
    """
    当前的机器人对象。
    """
    event_queue: List['T_Event'] = []
    """
    事件队列，用于 ``get()`` 方法。
    """
    max_event_queue_len: int = 1024
    """
    最大事件队列长度。
    """
    wait_for_get: bool = False
    """
    当此属性为 ``Ture`` 时，``handle_event()`` 将不对当前事件进行传播和处理，用于 ``get()`` 方法。
    """

    def __init__(self, bot: 'Bot'):
        self.bot: 'Bot' = bot

    async def safe_run(self):
        """
        附带有异常处理地安全运行适配器。
        """
        try:
            await self.run()
        except Exception as e:
            logger.error(f'Run adapter {self.__class__.__name__} failed: {e!r}')

    @abstractmethod
    async def run(self):
        """
        适配器运行方法，适配器开发者必须实现该方法。
        适配器运行过程中保持保持运行，当此方法结束后，AliceBot 不会自动重新启动适配器。
        """
        pass

    async def startup(self):
        """
        在适配器开始运行前运行的方法。
        AliceBot 依次运行并等待所有适配器的 ``startup()`` 方法，待运行完毕后再创建 ``run()`` 任务。
        """
        pass

    async def shutdown(self):
        """
        在适配器结束后运行后运行的方法。
        AliceBot 在接收到系统的结束信号后依次运行并等待所有适配器的 ``shutdown()`` 方法。当强制退出时此方法可能未被执行。
        """
        pass

    def handle_event(self, event: 'T_Event'):
        """
        调用 ``Bot`` 对象进行事件处理，并维护一个事件队列用于 ``get()`` 方法，所有适配器在接收到事件后必须使用此方法进行事件处理。
        :param event: 待处理的事件。
        """
        logger.info(f'Adapter {self.__class__.__name__} received: {event!r}')
        self.event_queue.append(event)
        if len(self.event_queue) > self.max_event_queue_len:
            self.event_queue.pop(0)
        if self.wait_for_get:
            self.wait_for_get = False
        else:
            event.handled = True
            self.bot.loop.create_task(self.bot.handle_event(event))

    async def get(self, current_event: 'T_Event',
                  func: Optional[Callable[['T_Event'], Union[bool, Awaitable[bool]]]] = None,
                  max_try_times: Optional[int] = None,
                  timeout: Optional[Union[int, float]] = None) -> Optional['T_Event']:
        """
        获取满足指定条件的的事件，协程会等待直到适配器接收到满足条件的事件、超过最大事件数或超时。
        当适配器接收到超过最大消息数的事件后仍未满足 func 的条件时返回 ``None`` 。

        :param current_event: 当前事件，方法将仅检索此事件后接收到的事件。
        :param func: (optional) 协程或者函数，函数会被自动包装为协程执行。要求接受一个事件作为参数，返回布尔值。当协程返回 ``True`` 时返回当前事件。
            当为 ``None`` 时相当于输入对于任何事件均返回真的协程，即返回适配器接收到的下一个事件。
        :param max_try_times: 最大事件数。
        :param timeout: 超时。
        :return: 返回满足 func 条件的事件。
        :rtype: Optional['T_Event']
        """

        async def always_true(_):
            return True

        def run_sync(_func):
            @wraps(_func)
            async def _wrapper(*args, **kwargs):
                return await self.bot.loop.run_in_executor(None, partial(_func, *args, **kwargs))

            return _wrapper

        if func is None:
            func = always_true
        elif not asyncio.iscoroutinefunction(func):
            func = run_sync(func)

        try_times = 0
        start_time = time.time()
        while not self.bot.should_exit and (max_try_times is None or (try_times < max_try_times)) and (
                timeout is None or (time.time() - start_time < timeout)):
            try:
                index = self.event_queue.index(current_event)
            except ValueError:
                raise ValueError('can not find current_event')
            for event in self.event_queue[index + 1:]:
                if not event.handled:
                    if await func(event):
                        event.handled = True
                        return event
                    else:
                        self.handle_event(event)
            self.wait_for_get = True
            while self.wait_for_get and not self.bot.should_exit:
                await asyncio.sleep(0)
            try_times += 1

        return None



"""
============
CQHTTP 配置
============
"""
from pydantic import BaseModel


class Config(BaseModel):
    """
    CQHTTP 配置类，将在适配器被加载时被混入到机器人主配置中。
    """
    __config_name__ = 'cqhttp'
    """
    配置名称。
    """
    host: str = '127.0.0.1'
    """
    本机域名。
    """
    port: int = 8080
    """
    监听的端口。
    """
    url: str = '/cqhttp/ws'
    """
    WebSockets 路径，需和客户端配置相同。
    """
    api_timeout: int = 1000
    """
    进行 API 调用时等待返回响应的超时时间。
    """



"""
======
配置
======
AliceBot 使用 pydantic_ 来读取配置。

.. _pydantic: https://pydantic-docs.helpmanual.io/
"""
from typing import Set, Optional

from pydantic import BaseModel

config_file = 'config.json'


class MainConfig(BaseModel):
    """
    AliceBot 的主要配置。
    """
    plugins: Optional[Set[str]]
    """
    将被加载的插件列表，将依次被 ``Bot`` 类的 ``load_plugin()`` 方法加载。
    
    :type: Optional[Set[str]]
    """
    plugin_dir: Optional[Set[str]]
    """
    将被加载的插件目录列表，将被 ``Bot`` 类的 ``load_plugins_from_dir()`` 方法加载。
    
    :type: Optional[Set[str]]
    """
    adapters: Optional[Set[str]]
    """
    将被加载的适配器列表，将依次被 ``Bot`` 类的 ``load_adapter()`` 方法加载。
    
    :type: Optional[Set[str]]
    """

    class Config:
        extra = "allow"



"""
============
CQHTTP 消息
============
"""
from typing import Any, Literal, Optional, TypeVar, Union, Mapping, Iterable

from alicebot.message import Message, MessageSegment

T_CQHTTPMessage = TypeVar('T_CQHTTPMessage', bound='CQHTTPMessage')
T_CQHTTPMessageSegment = TypeVar('T_CQHTTPMessageSegment', bound='CQHTTPMessageSegment')


class CQHTTPMessage(Message):
    """CQHTTP 消息"""

    @staticmethod
    def _construct(msg: Union[str, Mapping, Iterable[Mapping], Any]) -> Iterable[T_CQHTTPMessageSegment]:
        if isinstance(msg, Mapping):
            yield CQHTTPMessageSegment(msg['type'], msg.get('data') or {})
        elif isinstance(msg, Iterable) and not isinstance(msg, str):
            for seg in msg:
                yield CQHTTPMessageSegment(seg['type'], seg.get('data') or {})
        elif isinstance(msg, str):
            yield CQHTTPMessageSegment.text(msg)
        return

    def is_text(self) -> bool:
        """
        :return: 是否是纯文本消息。
        """
        for ms in self:
            if not ms.is_text():
                return False
        return True

    def get_plain_text(self) -> str:
        """
        获取消息中的纯文本部分。

        :return: 消息中的纯文本部分。
        :rtype: str
        """
        return ''.join(map(lambda x: str(x), filter(lambda x: x.is_text(), self)))


class CQHTTPMessageSegment(MessageSegment):
    """CQHTTP 消息字段"""

    def __str__(self) -> str:
        if self.type == 'text':
            return self.data.get('text', '')
        return self.get_cqcode()

    def __add__(self, other) -> T_CQHTTPMessage:
        return CQHTTPMessage(self) + other

    def __radd__(self, other) -> T_CQHTTPMessage:
        return CQHTTPMessage(other) + self

    def is_text(self) -> bool:
        """
        :return: 是否是纯文本消息字段。
        :rtype: bool
        """
        return self.type == 'text'

    def get_cqcode(self) -> str:
        """
        :return: 此消息字段的 CQ 码形式。
        :rtype: str
        """
        if self.type == 'text':
            return escape(self.data.get('text', ''), escape_comma=False)

        params = ','.join([f'{k}={escape(str(v))}' for k, v in self.data.items() if v is not None])
        return f'[CQ:{self.type}{"," if params else ""}{params}]'

    @classmethod
    def text(cls, text: str) -> T_CQHTTPMessageSegment:
        """纯文本"""
        return cls(**{
            "type": "text",
            "data": {
                "text": text
            }
        })

    @classmethod
    def face(cls, id_: int) -> T_CQHTTPMessageSegment:
        """QQ 表情"""
        return cls(**{
            "type": "face",
            "data": {
                "id": str(id_)
            }
        })

    @classmethod
    def image(cls,
              file: str,
              type_: Optional[Literal['flash']] = None,
              cache: bool = True,
              proxy: bool = True,
              timeout: Optional[int] = None) -> T_CQHTTPMessageSegment:
        """图片"""
        return cls(**{
            "type": "image",
            "data": {
                "file": file,
                "type": type_,
                "cache": cache,
                "proxy": proxy,
                "timeout": timeout
            }
        })

    @classmethod
    def record(cls,
               file: str,
               magic: bool = False,
               cache: bool = True,
               proxy: bool = True,
               timeout: Optional[int] = None) -> T_CQHTTPMessageSegment:
        """语音"""
        return cls(**{
            "type": "record",
            "data": {
                "file": file,
                "magic": magic,
                "cache": cache,
                "proxy": proxy,
                "timeout": timeout
            }
        })

    @classmethod
    def video(cls,
              file: str,
              cache: bool = True,
              proxy: bool = True,
              timeout: Optional[int] = None) -> T_CQHTTPMessageSegment:
        """短视频"""
        return cls(**{
            "type": "video",
            "data": {
                "file": file,
                "cache": cache,
                "proxy": proxy,
                "timeout": timeout
            }
        })

    @classmethod
    def at(cls, qq: Union[int, Literal['all']]) -> T_CQHTTPMessageSegment:
        """@某人"""
        return cls(**{
            "type": "at",
            "data": {
                "qq": str(qq)
            }
        })

    @classmethod
    def rps(cls) -> T_CQHTTPMessageSegment:
        """猜拳魔法表情"""
        return cls(**{
            "type": "rps",
            "data": {}
        })

    @classmethod
    def dice(cls) -> T_CQHTTPMessageSegment:
        """掷骰子魔法表情"""
        return cls(**{
            "type": "dice",
            "data": {}
        })

    @classmethod
    def shake(cls) -> T_CQHTTPMessageSegment:
        """窗口抖动（戳一戳）"""
        return cls(**{
            "type": "shake",
            "data": {}
        })

    @classmethod
    def poke(cls, type_: str, id_: int) -> T_CQHTTPMessageSegment:
        """戳一戳"""
        return cls(**{
            "type": "poke",
            "data": {
                "type": type_,
                "id": str(id_)
            }
        })

    @classmethod
    def anonymous(cls, ignore: Optional[bool] = None) -> T_CQHTTPMessageSegment:
        """匿名发消息"""
        return cls(**{
            "type": "anonymous",
            "data": {
                "ignore": ignore
            }
        })

    @classmethod
    def share(cls,
              url: str,
              title: str,
              content: Optional[str] = None,
              image: Optional[str] = None) -> T_CQHTTPMessageSegment:
        """链接分享"""
        return cls(**{
            "type": "share",
            "data": {
                "url": url,
                "title": title,
                "content": content,
                "image": image
            }
        })

    @classmethod
    def contact(cls, type_: Literal['qq', 'group'], id_: int) -> T_CQHTTPMessageSegment:
        """推荐好友/推荐群"""
        return cls(**{
            "type": "contact",
            "data": {
                "type": type_,
                "id": str(id_)
            }
        })

    @classmethod
    def contact_friend(cls, id_: int) -> T_CQHTTPMessageSegment:
        """推荐好友"""
        return cls(**{
            "type": "contact",
            "data": {
                "type": 'qq',
                "id": str(id_)
            }
        })

    @classmethod
    def contact_group(cls, id_: int) -> T_CQHTTPMessageSegment:
        """推荐好友"""
        return cls(**{
            "type": "contact",
            "data": {
                "type": 'group',
                "id": str(id_)
            }
        })

    @classmethod
    def location(cls,
                 lat: float,
                 lon: float,
                 title: Optional[str],
                 content: Optional[str] = None) -> T_CQHTTPMessageSegment:
        """位置"""
        return cls(**{
            "type": "location",
            "data": {
                "lat": str(lat),
                "lon": str(lon),
                "title": title,
                "content": content
            }
        })

    @classmethod
    def music(cls, type_: Literal['qq', '163', 'xm'], id_: int) -> T_CQHTTPMessageSegment:
        """音乐分享"""
        return cls(**{
            "type": "music",
            "data": {
                "type": type_,
                "id": str(id_)
            }
        })

    @classmethod
    def music_custom(cls,
                     url: str,
                     audio: str,
                     title: str,
                     content: Optional[str] = None,
                     image: Optional[str] = None) -> T_CQHTTPMessageSegment:
        """音乐自定义分享"""
        return cls(**{
            "type": "music",
            "data": {
                "type": "custom",
                "url": url,
                "audio": audio,
                "title": title,
                "content": content,
                "image": image
            }
        })

    @classmethod
    def reply(cls, id_: int) -> T_CQHTTPMessageSegment:
        """回复"""
        return cls(**{
            "type": "reply",
            "data": {
                "id": str(id_)
            }
        })

    @classmethod
    def node(cls, id_: int) -> T_CQHTTPMessageSegment:
        """合并转发节点"""
        return cls(**{
            "type": "node",
            "data": {
                "id": str(id_)
            }
        })

    @classmethod
    def node_custom(cls, user_id: int, nickname, content: T_CQHTTPMessage) -> T_CQHTTPMessageSegment:
        """合并转发自定义节点"""
        return cls(**{
            "type": "node",
            "data": {
                "user_id": str(user_id),
                "nickname": str(nickname),
                "content": content
            }
        })

    @classmethod
    def xml_message(cls, data: str) -> T_CQHTTPMessageSegment:
        """XML 消息"""
        return cls(**{
            "type": "xml",
            "data": {
                "data": data
            }
        })

    @classmethod
    def json_message(cls, data: str) -> T_CQHTTPMessageSegment:
        """JSON 消息"""
        return cls(**{
            "type": "json",
            "data": {
                "data": data
            }
        })


def escape(s: str, *, escape_comma: bool = True) -> str:
    """
    对 CQ 码中的特殊字符进行转义。

    :param s: 待转义的字符串。
    :param escape_comma: 是否转义 ``,``。
    :return: 转义后的字符串。
    :rtype: str
    """
    s = s.replace("&", "&amp;") \
        .replace("[", "&#91;") \
        .replace("]", "&#93;")
    if escape_comma:
        s = s.replace(",", "&#44;")
    return s



"""
======
事件
======
事件类的基类。适配器开发者应实现此事件类基类的子类。
"""
from abc import ABC
from typing import Any, Optional, TypeVar

from pydantic import BaseModel

from alicebot.message import Message, DataclassEncoder

T_Event = TypeVar('T_Event', bound='Event')


class Event(ABC, BaseModel):
    """
    事件类的基类。
    """
    adapter: Any
    """
    产生当前事件的适配器对象。
    """
    type: Optional[str]
    """
    事件类型。
    """
    handled: bool = False
    """
    表示事件是否被处理过了，用于适配器处理。
    注意：请勿手动更改此属性的值。
    """

    def __str__(self) -> str:
        return f'Event<{self.type}>'

    def __repr__(self) -> str:
        return self.__str__()

    class Config:
        extra = "allow"
        json_encoders = {Message: DataclassEncoder}



import inspect
import pkgutil
import importlib
from importlib.abc import MetaPathFinder
from typing import Iterable, List, Tuple, Type, TypeVar, Optional, TYPE_CHECKING
from importlib.machinery import PathFinder

from pydantic import BaseModel

from alicebot.log import logger
from alicebot.exception import LoadModuleError

if TYPE_CHECKING:
    from pkgutil import ModuleInfo

_T = TypeVar('_T')


class ModulePathFinder(MetaPathFinder):
    """
    用于查找 AliceBot 组件的元路径查找器。
    """
    path: List[str] = []

    def find_spec(self, fullname, path=None, target=None):
        if path is None:
            path = []
        return PathFinder.find_spec(fullname, self.path + path, target)


def load_module(name: str,
                module_class_type: Type[_T],
                *args, **kwargs) -> Tuple[Type[_T], Optional[Type[BaseModel]]]:
    """
    从模块中查找指定类型的类和 ``Config`` 。
    查找到指定的类后会尝试使用指定参数示例化类，仅返回能被实例化的类。
    若模块中存在多个符合条件的类，则抛出错误。

    :param name: 模块名称，格式和 Python ``import`` 语句相同。
    :param module_class_type: 要查找的类型。
    :param args: 实例化类时使用的参数。
    :param kwargs: 实例化类时使用的参数。
    :return: ``(class, config)`` 返回符合条件的类和配置类组成的元组。
    :rtype: Tuple[Type[_T], Optional[Type[BaseModel]]]
    :exception LoadModuleError: 当找不到符合条件的类或者模块中存在多个符合条件的类。
    """

    def _is_instantiate_class(_module_class: type):
        try:
            _module_class = _module_class(*args, **kwargs)
        except TypeError:
            return False
        else:
            return True

    importlib.invalidate_caches()
    module = importlib.import_module(name)

    module_class = []
    for module_attr in dir(module):
        module_attr = getattr(module, module_attr)
        if inspect.isclass(module_attr) and issubclass(module_attr, module_class_type):
            module_class.append(module_attr)

    module_class = list(filter(_is_instantiate_class, module_class))

    if not module_class:
        logger.error(f'Can not find {module_class_type!r} class in the {name!r} module')
    elif len(module_class) > 1:
        logger.error(f'More then one {module_class_type!r} class in the {name!r} module')
    else:
        if 'Config' in dir(module):
            module_attr = getattr(module, 'Config')
            if inspect.isclass(module_attr) and issubclass(module_attr, BaseModel) and '__config_name__' in dir(
                    module_attr):
                return module_class[0], getattr(module, 'Config')
        return module_class[0], None

    raise LoadModuleError


def load_modules_from_dir(_module_path_finder: ModulePathFinder,
                          path: Iterable[str],
                          module_class_type: Type[_T],
                          *args, **kwargs) -> List[Tuple[Type[_T], Optional[Type[BaseModel]], 'ModuleInfo']]:
    """
    从指定路径列表中的所有模块中查找指定类型的类和 ``Config`` ，以 ``_`` 开头的插件不会被导入。
    路径可以是相对路径或绝对路径。

    :param _module_path_finder: 用于查找 AliceBot 组件的元路径查找器。
    :param path: 由储存模块的路径文本组成的列表。 ``['path/of/plugins/', '/home/xxx/alicebot/plugins']``
    :param module_class_type: 要查找的类型。
    :param args: 实例化类时使用的参数。
    :param kwargs: 实例化类时使用的参数。
    :return: ``[(class, config, module_info), ...]`` 返回由符合条件的类、配置类和 ``ModuleInfo`` 组成的元组组成的列表。
    """
    modules = []
    _module_path_finder.path = list(path)
    for module_info in pkgutil.iter_modules(path):
        try:
            module_class, config_class = load_module(module_info.name, module_class_type, *args, **kwargs)
            if not module_info.name.startswith('_'):
                modules.append((module_class, config_class, module_info))
        except Exception:
            continue
    return modules



"""
======
异常
======
下列是 AliceBot 运行过程中可能会抛出的异常。这些异常大部分不需要用户处理，AliceBot 会自动捕获并处理。
对于适配器开发者，所有适配器抛出的异常都应该继承自 ``AdapterException`` 。
"""


class AliceBotException(Exception):
    """
    所有 AliceBot 发生的异常的基类。
    """
    pass


class EventException(AliceBotException):
    """
    事件处理过程中由插件抛出的异常，用于控制事件的传播，会被 AliceBot 自动捕获并处理。
    """
    pass


class SkipException(EventException):
    """
    跳过当前插件继续当前事件传播。
    """
    pass


class StopException(EventException):
    """
    停止当前事件传播。
    """
    pass


class AdapterException(AliceBotException):
    """
    由适配器抛出的异常基类，所有适配器抛出的异常都应该继承自此类。
    """
    pass


class LoadModuleError(AliceBotException):
    """
    加载模块错误，在指定模块中找不到特定类型的类或模块中存在多个符合条件的类时抛出。
    """
    pass



"""
==================
CQHTTP 协议适配器
==================
本适配器适配了 OneBot v11 协议。
协议详情请参考: `OneBot`_ 。

.. _OneBot: https://github.com/howmanybots/onebot/blob/master/README.md
"""
import sys
import time
import json
import asyncio
from functools import partial
from typing import Any, Dict, Iterable, List, Literal, Optional, Union, Mapping, TYPE_CHECKING

import aiohttp
from aiohttp import web

from alicebot.log import logger
from alicebot.adapter import AbstractAdapter
from alicebot.message import DataclassEncoder

from .config import Config
from .event import get_event_class
from .message import CQHTTPMessage, CQHTTPMessageSegment
from .exception import NetworkError, ActionFailed, ApiNotAvailable, ApiTimeout

if TYPE_CHECKING:
    from .message import T_CQHTTPMessage, T_CQHTTPMessageSegment


class CQHTTPAdapter(AbstractAdapter):
    """
    CQHTTP 协议适配器。
    在插件中可以直接使用 ``self.adapter.xxx_api(**params)`` 调用名称为 ``xxx_api`` 的 API，和调用 ``call_api()`` 方法相同。
    """
    name: str = 'cqhttp'
    app: web.Application
    runner: web.AppRunner
    site: web.TCPSite
    websocket: web.WebSocketResponse
    api_response: List[Dict[str, Any]] = []
    wait_for_get_api_response: bool = False
    _api_id: int = 0

    @property
    def config(self):
        """
        :return: 本适配器的配置。
        """
        return getattr(self.bot.config, Config.__config_name__)

    def __getattr__(self, item):
        return partial(self.call_api, item)

    async def startup(self):
        """
        创建 aiohttp Application。
        """
        self.app = web.Application()
        self.app.add_routes([web.get(self.config.url, self.handle_response)])

    async def run(self):
        """
        运行 aiohttp WebSockets 服务器。
        """
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.config.host, self.config.port)
        await self.site.start()

    async def shutdown(self):
        """
        清理 aiohttp AppRunner。
        """
        await self.runner.cleanup()

    async def handle_response(self, request: web.Request):
        """
        处理 aiohttp WebSockets 服务器的接收。

        :param request: aiohttp WebSockets 服务器的 Request 对象。
        """
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        self.websocket = ws

        msg: aiohttp.WSMessage
        async for msg in ws:
            if self.bot.should_exit:
                break

            if msg.type == aiohttp.WSMsgType.TEXT:
                try:
                    msg_dict = msg.json()
                except json.JSONDecodeError as e:
                    logger.error(f'WebSocket message parsing error, not json: {e!r}')
                    continue

                if 'post_type' in msg_dict:
                    self.handle_cqhttp_event(msg_dict)
                else:
                    self.handle_api_api_response(msg_dict)

            elif msg.type == aiohttp.WSMsgType.ERROR:
                logger.error(f'WebSocket connection closed with exception {ws.exception()!r}')

        logger.warning(f'WebSocket connection closed!')

        return ws

    def _get_api_echo(self) -> int:
        self._api_id = (self._api_id + 1) % sys.maxsize
        return self._api_id

    def handle_cqhttp_event(self, msg: Dict[str, Any]):
        """
        处理 CQHTTP 事件。

        :param msg: 接收到的信息。
        """
        post_type = msg.get('post_type')
        event_type = msg.get(post_type + '_type')
        sub_type = msg.get('sub_type', None)
        event_class = get_event_class(post_type, event_type, sub_type)

        cqhttp_event = event_class(**msg)
        cqhttp_event.adapter = self

        if cqhttp_event.post_type == 'meta_event':
            # meta_event 不交由插件处理
            if cqhttp_event.meta_event_type == 'lifecycle' and cqhttp_event.sub_type == 'connect':
                logger.info(f'WebSocket connection from CQHTTP Bot {msg.get("self_id")} accepted!')
            elif cqhttp_event.meta_event_type == 'heartbeat':
                if cqhttp_event.status.good and cqhttp_event.status.online:
                    pass
                else:
                    logger.error(f'CQHTTP Bot status is not good: {cqhttp_event.status.dict()}')
        else:
            self.handle_event(cqhttp_event)

    def handle_api_api_response(self, msg: Dict[str, Any]):
        """
        处理 CQHTTP API 调用的响应内容。

        :param msg: 接收到的信息。
        """
        self.wait_for_get_api_response = False
        self.api_response.append(msg)

    async def call_api(self, api: str, **params) -> Optional[Dict[str, Any]]:
        """
        调用 CQHTTP API，协程会等待直到获得 API 响应。

        :param api: API 名称。
        :param params: API 参数。
        :return: API 响应中的 data 字段。
        :rtype: Optional[Dict[str, Any]]
        :exception NetworkError: 网络错误。
        :exception ApiNotAvailable: API 请求响应 404， API 不可用。
        :exception ActionFailed: API 请求响应 failed， API 操作失败。
        :exception ApiTimeout: API 请求响应超时。
        """
        api_echo = self._get_api_echo()
        try:
            await self.websocket.send_str(json.dumps({
                'action': api,
                'params': params,
                'echo': api_echo
            }, cls=DataclassEncoder))
        except Exception:
            raise NetworkError

        timeout = self.config.api_timeout
        start_time = time.time()
        self.wait_for_get_api_response = True
        while not self.bot.should_exit and (time.time() - start_time < timeout):
            while self.wait_for_get_api_response and not self.bot.should_exit:
                await asyncio.sleep(0)
            for index, resp in enumerate(self.api_response):
                if resp['echo'] == api_echo:
                    if resp.get('retcode') == 1404:
                        raise ApiNotAvailable(resp=resp)
                    if resp.get('status') == 'failed':
                        raise ActionFailed(resp=resp)
                    return self.api_response.pop(index).get('data')

        if not self.bot.should_exit:
            raise ApiTimeout

    async def send(self,
                   _message: Union[str, Mapping, Iterable[Mapping], 'T_CQHTTPMessageSegment', 'T_CQHTTPMessage'],
                   message_type: Literal['private', 'group'],
                   _id: int) -> Optional[Dict[str, Any]]:
        """
        发送消息，调用 send_private_msg 或 send_group_msg API 发送消息。

        :param _message: 消息内容，可以是 str, Mapping, Iterable[Mapping], 'T_CQHTTPMessageSegment', 'T_CQHTTPMessage'。
            将使用 ``CQHTTPMessage`` 进行封装。
        :param message_type: 消息类型。应该是 private 或者 group。
        :param _id: 发送对象的 ID ，QQ 号码或者群号码。
        :return: API 响应。
        :rtype: Optional[Dict[str, Any]]
        :exception TypeError: message_type 不是 'private' 或 'group'。
        :exception ...: 同 ``call_api()`` 方法。
        """
        if message_type == 'private':
            return await self.send_private_msg(user_id=_id, message=CQHTTPMessage(_message))
        elif message_type == 'group':
            return await self.send_group_msg(group_id=_id, message=CQHTTPMessage(_message))
        else:
            raise TypeError('message_type must be "private" or "group"')



import sys
import json
import signal
import asyncio
import threading
from itertools import chain
from typing import Any, List, Dict, Iterable, Tuple, Type, TypeVar, Optional, TYPE_CHECKING

from pydantic import create_model, ValidationError

from alicebot.log import logger
from alicebot.plugin import Plugin
from alicebot.adapter import AbstractAdapter
from alicebot.config import MainConfig, BaseModel, config_file
from alicebot.exception import StopException, SkipException, LoadModuleError
from alicebot.load_module import load_module, ModulePathFinder, load_modules_from_dir

if TYPE_CHECKING:
    from alicebot.event import T_Event
    from alicebot.plugin import T_Plugin
    from alicebot.adapter import T_Adapter

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)

_T = TypeVar('_T')


class Bot:
    """
    AliceBot 机器人对象，定义了机器人的基本方法，读取并储存配置 ``Config`` ，加载适配器 ``Adapter`` 和插件 ``Plugin``，并进行事件分发。
    """
    config = None
    config_json: Dict[str, Any] = {}
    loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
    should_exit: bool = False
    adapters: List['T_Adapter'] = []
    plugins_priority_dict: Dict[int, List[Type['T_Plugin']]] = {}
    _module_path_finder: ModulePathFinder = ModulePathFinder()
    _config_update_dict: Dict[str, Tuple[Type[BaseModel], Any]] = {}

    def __init__(self, config_file_: Optional[str] = None):
        """
        初始化 AliceBot ，读取配置文件，创建配置，加载适配器和插件。

        :param config_file_: (optional) 指定配置文件，如不指定使用默认的 ``config.json`` 。
        """
        if config_file_ is None:
            config_file_ = config_file
        sys.meta_path.insert(0, self._module_path_finder)
        try:
            with open(config_file_, 'r', encoding='utf8') as f:
                self.config_json = json.load(f)
                self.config = MainConfig(**self.config_json)
        except OSError as e:
            logger.warning(f'Can not open config file: {e!r}')
        except json.JSONDecodeError as e:
            logger.warning(f'Read config file failed: {e!r}')
        except ValidationError as e:
            logger.error(f'Config file error: {e!r}')
        except ValueError as e:
            logger.error(f'Read config file failed: {e!r}')

        if self.config is None:
            return

        if self.config.plugin_dir:
            self.load_plugins_from_dir(self.config.plugin_dir)
        if self.config.plugins:
            for _plugin in self.config.plugins:
                self.load_plugin(_plugin)
        if self.config.adapters:
            for _adapter in self.config.adapters:
                self.load_adapter(_adapter)

    @property
    def plugins(self) -> List['T_Plugin']:
        """
        :return: 当前已经加载的插件的列表。
        :rtype: List['T_Plugin']
        """
        return list(chain(*self.plugins_priority_dict.values()))

    def run(self):
        """
        运行 AliceBot ，监听并拦截系统退出信号，更新机器人配置。
        """
        logger.info('Running AliceBot...')
        # 监听并拦截系统退出信号，从而完成一些善后工作后再关闭程序
        if threading.current_thread() is threading.main_thread():
            # Signal 仅能在主线程中被处理。
            try:
                for sig in HANDLED_SIGNALS:
                    self.loop.add_signal_handler(sig, self.handle_exit)
            except NotImplementedError:
                # add_signal_handler 仅在 Unix 下可用，以下对于 Windows。
                for sig in HANDLED_SIGNALS:
                    signal.signal(sig, self.handle_exit)

        # 更新config，合并入来自Plugin和Adapter的Config
        if self._config_update_dict and self.config_json:
            self.config = create_model('Config', **self._config_update_dict, __base__=MainConfig)(**self.config_json)

        try:
            self.loop.run_until_complete(self._run())
        finally:
            try:
                self._cancel_all_tasks()
                self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            finally:
                asyncio.set_event_loop(None)
                self.loop.close()

    async def _run(self):
        try:
            for _adapter in self.adapters:
                try:
                    await _adapter.startup()
                except Exception as e:
                    logger.error(f'Startup _adapter {_adapter!r} failed: {e!r}')
            for _adapter in self.adapters:
                self.loop.create_task(_adapter.safe_run())
            await self.main_loop()
        finally:
            for _adapter in self.adapters:
                await _adapter.shutdown()

    async def main_loop(self):
        while not self.should_exit:
            await asyncio.sleep(0.1)

    def handle_exit(self):
        """
        当机器人收到退出信号时，根据情况进行处理。
        """
        logger.info(f'Stopping AliceBot...')
        if self.should_exit:
            logger.warning(f'Force Exit AliceBot...')
            sys.exit()
        else:
            self.should_exit = True

    def _cancel_all_tasks(self):
        to_cancel = asyncio.all_tasks(self.loop)
        if not to_cancel:
            return
        for task in to_cancel:
            task.cancel()
        self.loop.run_until_complete(asyncio.gather(*to_cancel, loop=self.loop, return_exceptions=True))
        for task in to_cancel:
            if task.cancelled():
                continue
            if task.exception() is not None:
                logger.error('Unhandled exception during event loop shutdown')

    async def handle_event(self, current_event: 'T_Event'):
        """
        被适配器对象调用，根据优先级分发事件给所有插件，并处理插件的 ``stop`` 、 ``skip`` 等信号。
        此方法不应该被用户手动调用。

        :param current_event: 当前待处理的 ``Event``。
        """
        if current_event is None:
            return
        for plugin_priority in sorted(self.plugins_priority_dict.keys()):
            try:
                logger.debug(f'Checking for matching plugins with priority {plugin_priority!r}')
                for _plugin in self.plugins_priority_dict[plugin_priority]:
                    try:
                        _plugin = _plugin(current_event)
                        if await _plugin.rule():
                            logger.info(f'Event will be handled by {_plugin!r}')
                            await _plugin.handle()
                            if _plugin.block:
                                raise StopException()
                    except SkipException or StopException as e:
                        raise e
                    except Exception as e:
                        logger.error(f'Exception in plugin "{_plugin}": {e!r}')
            except StopException:
                # 插件要求停止当前事件传播
                break
            except SkipException:
                # 插件要求跳过自身继续当前事件传播
                continue
            except Exception as e:
                logger.error(f'Exception in handling event {current_event!r}: {e!r}')
            logger.info(f'Event Finished')

    def _load_plugin(self, plugin_class: Type['T_Plugin']):
        if type(plugin_class.priority) is int and plugin_class.priority >= 0:
            if plugin_class.priority in self.plugins_priority_dict:
                self.plugins_priority_dict[plugin_class.priority].append(plugin_class)
            else:
                self.plugins_priority_dict[plugin_class.priority] = [plugin_class]
        else:
            logger.error(f'Plugin class priority incorrect in the module "{plugin_class!r}"')
            raise LoadModuleError()

    def load_plugin(self, name: str) -> Optional[Type['T_Plugin']]:
        """
        加载单个插件。

        :param name: 插件名称，格式和 Python ``import`` 语句相同，
        :return: 被加载的插件类。
        :rtype: Optional[Type['T_Plugin']]
        """
        try:
            plugin_class, config_class = load_module(name, Plugin, None)
            self._load_plugin(plugin_class)
        except Exception as e:
            logger.error(f'Import plugin "{name}" failed: {e!r}')
        else:
            self._update_config(config_class)
            logger.info(f'Succeeded to import plugin "{name}"')
            return plugin_class

    def load_adapter(self, name: str) -> Optional[Type['T_Adapter']]:
        """
        加载单个适配器。

        :param name: 适配器名称，格式和 Python ``import`` 语句相同，
        :return: 被加载的适配器类。
        :rtype: Optional[Type['T_Adapter']]
        """
        try:
            adapter_class, config_class = load_module(name, AbstractAdapter, self)
            self.adapters.append(adapter_class(self))
        except Exception as e:
            logger.error(f'Load adapter "{name}" failed: {e!r}')
        else:
            self._update_config(config_class)
            logger.info(f'Succeeded to load adapter "{name}"')
            return adapter_class

    def load_plugins_from_dir(self, path: Iterable[str]):
        """
        从指定路径列表中加载插件，以 ``_`` 开头的插件不会被导入。
        路径可以是相对路径或绝对路径。

        :param path: 由储存插件的路径文本组成的列表。 ``['path/of/plugins/', '/home/xxx/alicebot/plugins']``
        """
        for module, config_class, module_info in load_modules_from_dir(self._module_path_finder, path, Plugin, None):
            try:
                self._load_plugin(module)
            except Exception as e:
                logger.error(
                    f'Import plugin "{module_info.name}" from path "{module_info.module_finder.path}" failed: {e!r}')
            else:
                self._update_config(config_class)
                logger.info(
                    f'Succeeded to import plugin "{module_info.name}" from path "{module_info.module_finder.path}"')

    def _update_config(self, config_class: Optional[Type['BaseModel']]):
        def _get_default_value():
            try:
                _config = config_class()
            except ValidationError:
                return ...
            else:
                return _config

        if config_class is not None:
            self._config_update_dict[getattr(config_class, '__config_name__')] = (config_class, _get_default_value())



"""
======
日志
======
AliceBot 使用 `loguru`_ 来记录日志信息。
自定义 logger 请参考 `loguru`_ 文档。

.. _loguru: https://github.com/Delgan/loguru
"""
from loguru import logger as _logger

logger = _logger
# logger = logging.getLogger("AliceBot")



"""
======
插件
======
所有 AliceBot 插件的基类。所有用户编写的插件必须继承自 ``Plugin`` 类。
"""
from abc import ABC, abstractmethod
from typing import Awaitable, Callable, TypeVar, Union, Optional, TYPE_CHECKING

from alicebot.exception import StopException, SkipException

if TYPE_CHECKING:
    from alicebot import Bot
    from alicebot.event import T_Event
    from alicebot.adapter import T_Adapter

T_Plugin = TypeVar('T_Plugin', bound='Plugin')


class Plugin(ABC):
    """
    所有 AliceBot 插件的基类。
    """
    event: 'T_Event'
    """
    当前正在被此插件处理的事件。
    """
    priority: int = 0
    """
    插件的优先级，数字越小表示优先级越高，默认为 0。
    """
    block: bool = False
    """
    插件执行结束后是否阻止事件的传播。``True`` 表示阻止。
    相当于在 ``handle()`` 方法最后调用 ``self.stop()`` 。
    """

    def __init__(self, event):
        self.event = event

    @property
    def name(self) -> str:
        """
        :return: 插件类名称。
        :rtype: str
        """
        return self.__class__.__name__

    @property
    def adapter(self) -> 'T_Adapter':
        """
        :return: 产生当前事件的适配器对象。
        :rtype: T_Adapter
        """
        return self.event.adapter

    @property
    def bot(self) -> 'Bot':
        """
        :return: 机器人对象。
        """
        return self.adapter.bot

    def stop(self):
        """
        停止当前事件传播。
        """
        raise StopException()

    def skip(self):
        """
        跳过自身继续当前事件传播。
        """
        raise SkipException()

    async def send(self, *args, **kwargs):
        """
        发送消息，具体实现和参数取决于适配器。
        """
        return await self.adapter.send(*args, **kwargs)

    async def get(self,
                  func: Optional[Callable[['T_Event'], Union[bool, Awaitable[bool]]]] = None,
                  max_try_times: int = None,
                  timeout: Optional[Union[int, float]] = None) -> Optional['T_Event']:
        """
        获取满足指定条件的的事件，协程会等待直到适配器接收到满足条件的事件、超过最大事件数或超时。
        当适配器接收到超过最大消息数的事件后仍未满足 func 的条件时，返回 ``None`` 。

        :param func: (optional) 协程或者函数，函数会被自动包装为协程执行。要求接受一个事件作为参数，返回布尔值。当协程返回 ``True`` 时返回当前事件。
            当为 ``None`` 时相当于输入对于任何事件均返回真的协程，即返回适配器接收到的下一个事件。
        :param max_try_times: 最大事件数。
        :param timeout: 超时。
        :return: 返回满足 func 条件的事件。
        :rtype: Optional['T_Event']
        """
        return await self.adapter.get(self.event, func, max_try_times, timeout)

    @abstractmethod
    async def handle(self) -> None:
        """
        处理事件的方法。当 ``rule()`` 方法返回 ``True`` 时 AliceBot 会调用此方法。每个插件必须实现此方法。
        """
        pass

    @abstractmethod
    async def rule(self) -> bool:
        """
        匹配事件的方法。事件处理时，会按照插件的优先级依次调用此方法，当此方法返回 ``True`` 时将事件交由此插件处理。每个插件必须实现此方法。
        注意：不建议直接在此方法内实现对事件的处理，事件的具体处理请交由 ``handle()`` 方法。

        :rtype: bool
        """
        pass



"""
========================
HTTP 服务端适配器示例
========================
这里是一个最简单可以直接使用的 HTTP 服务端适配器示例。
"""
from typing import Union, TYPE_CHECKING

from aiohttp import web

from alicebot.event import Event
from alicebot.message import Message, MessageSegment
from alicebot.adapter.utils import HttpServerAdapter

if TYPE_CHECKING:
    from alicebot.message import T_Message, T_MessageSegment


class HttpServerTestEvent(Event):
    """
    HTTP 服务端示例适配器事件类。
    """
    pass


class HttpServerTestAdapter(HttpServerAdapter):
    """
    HTTP 服务端示例适配器类。
    """
    get_url: str = '/'
    post_url: str = '/'
    host: str = '127.0.0.1'
    port: int = 8080

    async def handle_response(self, request: web.Request):
        event = HttpServerTestEvent(adapter=self,
                                    type='message',
                                    message=Message(MessageSegment(type='text', data={'text': await request.text()})))
        self.handle_event(event)

        return web.Response()

    @staticmethod
    async def send(msg: Union[str, 'T_Message', 'T_MessageSegment']):
        print(msg)



"""
============
CQHTTP 异常
============
"""
from alicebot.exception import AdapterException


class CQHTTPException(AdapterException):
    """
    CQHTTP 异常基类。
    """
    pass


class NetworkError(CQHTTPException):
    """
    网络异常。
    """
    pass


class ActionFailed(CQHTTPException):
    """
    API 请求成功响应，当响应表示 API 操作失败。
    :param resp: 返回的响应。
    """

    def __init__(self, resp):
        self.resp = resp


class ApiNotAvailable(ActionFailed):
    """
    API 请求返回 404，表示当前请求的 API 不可用或不存在。
    """


class ApiTimeout(AdapterException):
    """
    API 请求响应超时。
    """
    pass



"""
============
CQHTTP 事件
============
"""
import inspect
from typing import Literal, Optional, Type, TypeVar, Union, Mapping, Iterable, TYPE_CHECKING

from pydantic import BaseModel, Field

from alicebot.event import Event

from .message import CQHTTPMessage

if TYPE_CHECKING:
    from .message import T_CQHTTPMessage, T_CQHTTPMessageSegment

T_CQHTTPEvent = TypeVar('T_CQHTTPEvent', bound='CQHTTPEvent')


class Sender(BaseModel):
    user_id: Optional[int]
    nickname: Optional[str]
    card: Optional[str]
    sex: Optional[Literal['male', 'female', 'unknown']]
    age: Optional[int]
    area: Optional[str]
    level: Optional[str]
    role: Optional[str]
    title: Optional[str]


class Anonymous(BaseModel):
    id: int
    name: str
    flag: str


class File(BaseModel):
    id: str
    name: str
    size: int
    busid: int


class Status(BaseModel):
    online: bool
    good: bool

    class Config:
        extra = "allow"


class CQHTTPEvent(Event):
    """CQHTTP 事件基类"""
    __event__ = ''
    type: Optional[str] = Field(alias='post_type')
    time: int
    self_id: int
    post_type: Literal['message', 'notice', 'request', 'meta_event']

    @property
    def to_me(self) -> bool:
        """
        :return: 当前事件的 user_id 是否等于 self_id。
        :rtype: bool
        """
        return getattr(self, 'user_id') == self.self_id


class MessageEvent(CQHTTPEvent):
    """消息事件"""
    __event__ = 'message'
    post_type: Literal['message']
    message_type: Literal['private', 'group']
    sub_type: str
    message_id: int
    user_id: int
    message: CQHTTPMessage
    raw_message: str
    font: int
    sender: Sender

    def __repr__(self) -> str:
        return f'Event<{self.type}>: "{self.message}"'

    def get_plain_text(self) -> str:
        """
        :return: 消息的纯文本内容。
        :rtype: str
        """
        return self.message.get_plain_text()

    async def replay(self, msg: Union[str, Mapping, Iterable[Mapping], 'T_CQHTTPMessageSegment', 'T_CQHTTPMessage']):
        """
        回复消息。

        :param msg: 回复消息的内容，同 ``call_api()`` 方法。
        :return: API 请求响应。
        """
        pass


class PrivateMessageEvent(MessageEvent):
    """私聊消息"""
    __event__ = 'message.private'
    message_type: Literal['private']
    sub_type: Literal['friend', 'group', 'other']

    async def replay(self, msg: Union[str, Mapping, Iterable[Mapping], 'T_CQHTTPMessageSegment', 'T_CQHTTPMessage']):
        return await self.adapter.send_private_msg(user_id=self.user_id, message=CQHTTPMessage(msg))


class GroupMessageEvent(MessageEvent):
    """群消息"""
    __event__ = 'message.group'
    message_type: Literal['group']
    sub_type: Literal['normal', 'anonymous', 'notice']
    group_id: int
    anonymous: Optional[Anonymous] = None

    async def replay(self, msg: Union[str, Mapping, Iterable[Mapping], 'T_CQHTTPMessageSegment', 'T_CQHTTPMessage']):
        return await self.adapter.send_group_msg(group_id=self.group_id, message=CQHTTPMessage(msg))


class NoticeEvent(CQHTTPEvent):
    """通知事件"""
    __event__ = 'notice'
    post_type: Literal['notice']
    notice_type: str


class GroupUploadNoticeEvent(NoticeEvent):
    """群文件上传"""
    __event__ = 'notice.group_upload'
    notice_type: Literal["group_upload"]
    user_id: int
    group_id: int
    file: File


class GroupAdminNoticeEvent(NoticeEvent):
    """群管理员变动"""
    __event__ = 'notice.group_admin'
    notice_type: Literal['group_admin']
    sub_type: Literal['set', 'unset']
    user_id: int
    group_id: int


class GroupDecreaseNoticeEvent(NoticeEvent):
    """群成员减少"""
    __event__ = 'notice.group_decrease'
    notice_type: Literal['group_decrease']
    sub_type: Literal['leave', 'kick', 'kick_me']
    group_id: int
    operator_id: int
    user_id: int


class GroupIncreaseNoticeEvent(NoticeEvent):
    """群成员增加"""
    __event__ = 'notice.group_increase'
    notice_type: Literal['group_increase']
    sub_type: Literal['approve', 'invite']
    group_id: int
    operator_id: int
    user_id: int


class GroupBanNoticeEvent(NoticeEvent):
    """群禁言"""
    __event__ = 'notice.group_ban'
    notice_type: Literal['group_ban']
    sub_type: Literal['ban', 'lift_ban']
    group_id: int
    operator_id: int
    user_id: int
    duration: int


class FriendAddNoticeEvent(NoticeEvent):
    """好友添加"""
    __event__ = 'notice.friend_add'
    notice_type: Literal['friend_add']
    user_id: int


class GroupRecallNoticeEvent(NoticeEvent):
    """群消息撤回"""
    __event__ = 'notice.group_recall'
    notice_type: Literal['group_recall']
    group_id: int
    operator_id: int
    user_id: int
    message_id: int


class FriendRecallNoticeEvent(NoticeEvent):
    """好友消息撤回"""
    __event__ = 'notice.friend_recall'
    notice_type: Literal['friend_recall']
    user_id: int
    message_id: int


class NotifyEvent(NoticeEvent):
    """提醒事件"""
    __event__ = 'notice.notify'
    notice_type: Literal['notify']
    sub_type: str
    group_id: int
    user_id: int


class PokeNotifyEvent(NotifyEvent):
    """戳一戳"""
    __event__ = 'notice.notify.poke'
    sub_type: Literal['poke']
    target_id: int
    group_id: Optional[int] = None


class GroupLuckyKingNotifyEvent(NotifyEvent):
    """群红包运气王"""
    __event__ = 'notice.notify.lucky_king'
    sub_type: Literal['lucky_king']
    target_id: int


class GroupHonorNotifyEvent(NotifyEvent):
    """群成员荣誉变更"""
    __event__ = 'notice.notify.honor'
    sub_type: Literal['honor']
    honor_type: Literal['talkative', 'performer', 'emotion']


class RequestEvent(CQHTTPEvent):
    """请求事件"""
    __event__ = 'request'
    post_type: Literal['request']
    request_type: str

    async def approve(self):
        pass

    async def refuse(self):
        pass


class FriendRequestEvent(RequestEvent):
    """加好友请求"""
    __event__ = 'request.friend'
    request_type: Literal['friend']
    user_id: int
    comment: str
    flag: str

    async def approve(self, remark: str = ''):
        """
        同意请求。

        :param remark: (optional) 好友备注。
        :return: API 请求响应。
        """
        return await self.adapter.set_friend_add_request(flag=self.flag, approve=True, remark=remark)

    async def refuse(self):
        """
        拒绝请求。

        :return: API 请求响应。
        """
        return await self.adapter.set_friend_add_request(flag=self.flag, approve=False)


class GroupRequestEvent(RequestEvent):
    """加群请求／邀请"""
    __event__ = 'request.group'
    request_type: Literal['group']
    sub_type: Literal['add', 'invite']
    group_id: int
    user_id: int
    comment: str
    flag: str

    async def approve(self):
        """
        同意请求。

        :return: API 请求响应。
        """
        return await self.adapter.set_group_add_request(flag=self.flag, sub_type=self.sub_type, approve=True)

    async def refuse(self, reason: str = ''):
        """
        拒绝请求。

        :param reason: (optional) 拒绝原因。
        :return: API 请求响应。
        """
        return await self.adapter.set_group_add_request(flag=self.flag,
                                                        sub_type=self.sub_type,
                                                        approve=False,
                                                        reason=reason)


class MetaEvent(CQHTTPEvent):
    """元事件"""
    __event__ = 'meta_event'
    post_type: Literal['meta_event']
    meta_event_type: str


class LifecycleMetaEvent(MetaEvent):
    """生命周期"""
    __event__ = 'meta_event.lifecycle'
    meta_event_type: Literal['lifecycle']
    sub_type: Literal['enable', 'disable', 'connect']


class HeartbeatMetaEvent(MetaEvent):
    """心跳"""
    __event__ = 'meta_event.heartbeat'
    meta_event_type: Literal['heartbeat']
    status: Status
    interval: int


_cqhttp_events = {}
# define `model` first to avoid globals changing while `for`
# thanks for NoneBot
model = None
for model in globals().values():
    if inspect.isclass(model) and issubclass(model, CQHTTPEvent):
        _cqhttp_events[model.__event__] = model


def get_event_class(post_type: str, event_type: str, sub_type: Optional[str] = None) -> Type[T_CQHTTPEvent]:
    """
    根据接收到的消息类型返回对应的事件类。
    :param post_type: 请求类型。
    :param event_type: 事件类型。
    :param sub_type: (optional) 子类型。
    :return: 返回事件类。
    :rtype: Type[T_CQHTTPEvent]
    """
    if sub_type is None:
        return _cqhttp_events['.'.join((post_type, event_type))]
    return _cqhttp_events.get(
        '.'.join((post_type, event_type, sub_type))) or _cqhttp_events['.'.join((post_type, event_type))]



"""
==================
适配器实用工具
==================
这里定义了一些在编写适配器时常用的基类，适配器开发者可以直接继承自这里的类或者用作参考。
"""
import asyncio
from typing import Union
from abc import ABCMeta, abstractmethod

import aiohttp
from aiohttp import web

from alicebot.adapter import AbstractAdapter


class PollingAdapter(AbstractAdapter, metaclass=ABCMeta):
    """
    轮询式适配器示例。
    """
    delay: Union[int, float] = 0.1
    create_task: bool = False

    async def run(self):
        while not self.bot.should_exit:
            await asyncio.sleep(self.delay)
            if self.create_task:
                asyncio.create_task(self.on_tick())
            else:
                await self.on_tick()

    @abstractmethod
    async def on_tick(self):
        pass


class HttpClientAdapter(PollingAdapter, metaclass=ABCMeta):
    """
    HTTP 客户端适配器示例。
    """
    session: aiohttp.ClientSession

    async def startup(self):
        self.session = aiohttp.ClientSession()

    @abstractmethod
    async def on_tick(self):
        pass

    async def shutdown(self):
        await self.session.close()


class WebSocketClientAdapter(AbstractAdapter, metaclass=ABCMeta):
    """
    WebSocket 客户端适配器示例。
    """
    url: str

    async def run(self):
        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(self.url) as ws:
                msg: aiohttp.WSMessage
                async for msg in ws:
                    if self.bot.should_exit:
                        break
                    if msg.type == aiohttp.WSMsgType.ERROR:
                        break
                    await self.handle_response(msg)

    @abstractmethod
    def handle_response(self, msg: aiohttp.WSMessage):
        pass


class HttpServerAdapter(AbstractAdapter, metaclass=ABCMeta):
    """
    HTTP 服务端适配器示例。
    """
    app: web.Application
    runner: web.AppRunner
    site: web.TCPSite
    host: str
    port: int
    get_url: str
    post_url: str

    async def startup(self):
        self.app = web.Application()
        self.app.add_routes([web.get(self.get_url, self.handle_response),
                             web.post(self.post_url, self.handle_response)])

    async def run(self):
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.host, self.port)
        await self.site.start()

    async def shutdown(self):
        await self.runner.cleanup()

    @abstractmethod
    async def handle_response(self, request: web.Request):
        pass


class WebSocketServerAdapter(AbstractAdapter, metaclass=ABCMeta):
    """
    WebSocket 服务端适配器示例。
    """
    app: web.Application
    runner: web.AppRunner
    site: web.TCPSite
    websocket: web.WebSocketResponse
    host: str
    port: int
    url: str

    async def startup(self):
        self.app = web.Application()
        self.app.add_routes([web.get(self.url, self.handle_response)])

    async def run(self):
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, self.host, self.port)
        await self.site.start()

    async def shutdown(self):
        await self.runner.cleanup()

    async def handle_response(self, request: web.Request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        self.websocket = ws

        msg: aiohttp.WSMessage
        async for msg in ws:
            if self.bot.should_exit:
                break
            if msg.type == aiohttp.WSMsgType.TEXT:
                await self.handle_ws_response(msg)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                break

        return ws

    @abstractmethod
    async def handle_ws_response(self, msg: aiohttp.WSMessage):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alicebot',
 'alicebot.adapter',
 'alicebot.adapter.cqhttp',
 'alicebot.event',
 'alicebot.message',
 'alicebot.plugin']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.0,<4.0.0', 'loguru>=0.5.0,<0.6.0', 'pydantic>=1.8.0,<1.9.0']

setup_kwargs = {
    'name': 'alicebot',
    'version': '0.1.0.post3',
    'description': 'A simply asynchronous python chatbot framework.',
    'long_description': '<div align="center">\n  <a href="https://docs.alicebot.dev/"><img src="https://raw.githubusercontent.com/st1020/alicebot/master/docs/.vuepress/public/logo.png" width="200" height="200" alt="logo"></a>\n\n# AliceBot\n\n**简单的 Python 异步多后端机器人框架**\n\n</div>\n\n<div align="center">\n  <a href="https://raw.githubusercontent.com/st1020/alicebot/master/LICENSE">\n    <img src="https://img.shields.io/github/license/st1020/alicebot" alt="license">\n  </a>\n  <a href="https://pypi.python.org/pypi/alicebot">\n    <img src="https://img.shields.io/pypi/v/alicebot" alt="pypi">\n  </a>\n  <a href="https://pypi.python.org/pypi/alicebot">\n    <img src="https://img.shields.io/pypi/pyversions/alicebot" alt="pypi">\n  </a>\n  <a href="https://github.com/st1020/alicebot/">\n    <img src="https://img.shields.io/github/stars/st1020/alicebot?style=social" alt="github">\n  </a>\n</div>\n\n<p align="center">\n  <a href="https://docs.alicebot.dev/">文档</a>\n  ·\n  <a href="https://docs.alicebot.dev/guide/">指南</a>\n  ·\n  <a href="https://docs.alicebot.dev/guide/">API 参考</a>\n</p>\n\n## 简介\n\nAliceBot 是一个简单的 Python 异步多后端机器人框架，支持多种协议适配，可以轻松地编写易用学习和使用的插件来拓展其功能。\n\n本项目受到了 [NoneBot](https://github.com/nonebot/nonebot2/) 项目的启发，您可以在 [Why AliceBot?](#Why AliceBot?) 小节中查看这两个项目的异同，以便您选择更适合您的机器人框架。\n\n## 特点\n\nAliceBot 使用了非常灵活且易于使用的插件编写方式，您只需要编写两个方法即可实现一个功能强大的插件。\n\nAliceBot 的适配协议并不和任何一种库或网络协议绑定，您可以自由选择或编写适合您的适配器。\n\n目前 AliceBot 官方维护了以下协议适配：\n\n- 基于 WebSocket 的 [OneBot (CQHTTP) 协议](https://github.com/nonebot/nonebot2/blob/master/README.md) （支持QQ等）\n\n更多协议正在适配中 ...\n\n## 即刻开始\n\n1. 安装：\n\n   ```bash\n   pip install alicebot\n   ```\n\n2. 第一个 AliceBot 项目：\n\n   ```python\n   from alicebot import Bot\n   \n   bot = Bot()\n   bot.load_adapter(\'alicebot.adapter.cqhttp\')\n   \n   bot.run()\n   ```\n\n3. 第一个 AliceBot 插件：\n\n   ```python\n   from alicebot.plugin import Plugin\n   \n   \n   class Echo(Plugin):\n       async def handle(self) -> None:\n           await self.event.replay(self.event.message.replace(\'echo \', \'\'))\n   \n       async def rule(self) -> bool:\n           return self.adapter.name == \'cqhttp\' and self.event.type == \'message\' and self.event.message.startswith(\'echo \')\n   ```\n\n更多信息请参阅 AliceBot [文档](https://docs.alicebot.dev/)。\n\n## Why AliceBot?\n\n本项目受到了 [NoneBot](https://github.com/nonebot/nonebot2/) 项目的启发，以下介绍两者的异同。\n\n相同点：\n\n- 两者都是使用 Python 编写的，使用了协程异步的高性能机器人框架。\n- 两者都支持多种协议。\n- 两者都会对机器人收到的事件进行解析和处理，并按优先级分发给插件（事件响应器）来完成具体的功能。\n- 两者都基于 MIT 协议开源，这意味着您可以在遵循协议的前提下任意使用本项目。\n\n不同点：\n\n- 总的来说，NoneBot 是一个较为全面的机器人框架，而 AliceBot 是一个小巧简洁的机器人框架，它不包含一些复杂的高级特性，但更加灵活且易于学习（作者的观点）。\n- NoneBot 在实现上与 WebSocket 通讯协议深度绑定，它需要一个支持 ASGI 服务器协议的驱动框架，而 AliceBot 并不与任何协议绑定，它甚至可以用来驱动您的树莓派智能音箱。当然，如果您只需要一个支持常见网络聊天工具的机器人框架的话，它们并没有什么区别。\n- NoneBot 提供了很多预设规则和权限控制机制，而 AliceBot 则没有提供，由于插件编写方式的不同，AliceBot 选择给您最大的自由，您需要自行编写插件在何时运行的方法（`rule()` 方法），这在绝大部份情况下并不会造成您编写更多的代码或是影响插件的可读性，自行实现也不会非常困难，但如果您真的需要这些高级功能又不想自己实现的话，NoneBot 提供了更好的选择。\n- NoneBot 拥有相对庞大的用户基数和社区规模，也拥有数量众多的插件，而 AliceBot 则是一个新生项目，这意味着如果您使用 NoneBot 您可能会更加容易找到已经编写完毕的您感兴趣的插件，并且您当您遇到问题时也能够更快地查找到相关资料或者获得解答。\n- NoneBot 相对较为庞大，封装相对较多，完全学习理解需要一定时间，而 AliceBot 小巧简洁，封装较少，很容易掌握。\n\n总而言之，两者有着各自的特点，您可以根据需要选用。\n\n作者就是在使用 NoneBot 时觉得插件编写不是很易用，于是萌发了编写一个新的机器人框架的想法，AliceBot 就这样诞生了。\n\n## 协议与致谢\n\nAliceBot 采用 MIT 许可证开放源代码。\n\nAliceBot 的 message 模块和 cqhttp 适配器部分参考了 NoneBot 的实现，在此感谢 [NoneBot](https://github.com/nonebot/nonebot2/) 项目。',
    'author': 'st1020',
    'author_email': 'stone_1020@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://docs.alicebot.dev/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
